/*
 *   ScenarLang a domain specific language that attempts to make it even easier
 *   to write unit tests with mockito
 *    
 *   Copyright (C) 2012  X. Gillard 
 *   
 *   This library is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU Lesser General Public
 *   License as published by the Free Software Foundation; either
 *   version 2.1 of the License, or (at your option) any later version.
 *   
 *   This library is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *   Lesser General Public License for more details.
 *   
 *   You should have received a copy of the GNU Lesser General Public
 *   License along with this library; if not, write to the Free Software
 *   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 */
package xgi.ut.dsl.junit.rule;

import java.io.File;
import java.lang.reflect.Field;
import java.net.URISyntaxException;
import java.net.URL;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashMap;
import java.util.Map;

import org.junit.rules.MethodRule;
import org.junit.runners.model.FrameworkMethod;
import org.junit.runners.model.Statement;

import scala.collection.JavaConversions;
import xgi.ut.dsl.interpreters.ext.ScenarioInterpretation;
import xgi.ut.dsl.junit.annot.Context;
import xgi.ut.dsl.junit.annot.Scenario;
import xgi.ut.dsl.utl.ReflectionUtil;

/**
 * This is the junit method rule that manages the behavior of the
 * method that are annotated with @Scenario.
 * 
 * Basically, it will parse the scenario file pointed to by
 * the method's scenario annotation, interpret it, 
 * and then give you access to the declared variables.
 */
@SuppressWarnings("deprecation")
public class ScenarioRule implements MethodRule {
	private static final String BLANK = "\\s*";
	/**
	 * The current interpretation of the scenario that is associated with the current method
	 */
	private ScenarioInterpretation currentInterpretation = null;
	
	/**
	 * Basically, it will parse the scenario file pointed to by
	 * the method's scenario annotation, interpret it, 
	 * and then give you access to the declared variables.
	 * 
	 * {@inheritDoc}
	 */
	public Statement apply(final Statement base, final FrameworkMethod method, final Object target) {
		return new Statement() {
			@Override
			public void evaluate() throws Throwable {
				currentInterpretation = null;
				Scenario scenarioAnnot = method.getAnnotation(Scenario.class);
				if(scenarioAnnot==null){
					base.evaluate();
				} else {
					System.out.println(String.format("Initializing test '%s' with '%s'",
							method.getName(), scenarioAnnot.description()));
					
					// begin ensure backward compat with 1.1.8
					ArrayList<String> resourcesToLoad = new ArrayList<String>();
					if(!scenarioAnnot.scriptFile().matches(BLANK)){
						resourcesToLoad.add(scenarioAnnot.scriptFile());
					}
					resourcesToLoad.addAll(Arrays.asList(scenarioAnnot.scripts()));
					// end ensure backaward compat with 1.1.8
					
					currentInterpretation = new ScenarioInterpretation(getContext(target), getFiles(resourcesToLoad, target), new HashMap<String, Object>());
					base.evaluate();
				}
			}
		};
	}

	/**
	 * @return the variables that have been interpreted by the current scenario
	 */
	public Map<String, ?> getVariables(){
		if(currentInterpretation == null){
			return new HashMap<String, Object>();
		} else {
			return JavaConversions.mutableMapAsJavaMap(currentInterpretation.variables());
		}
	}
	/**
	 * @param target the testcase that may contain a field annotated with @Context
	 * @return return the field that is annotated with @Context or the testcase itself
	 */
	private Object getContext(Object target){
		for(Field f: ReflectionUtil.getAllFields(target.getClass()).values()){
			if(f.isAnnotationPresent(Context.class)){
				return f;
			}
		}
		return target;
	}
	
	/**
	 * @param fnames the name of the files to load
	 * @param target the object on which this rule apply
	 * @return the array of corresponding files
	 * @throws URISyntaxException 
	 */
	private File[] getFiles(ArrayList<String> fnames, Object target) throws URISyntaxException {
		File[] ret = new File[fnames.size()];
		for(int i = 0; i<fnames.size(); i++){
			URL scriptURL = target.getClass().getResource(fnames.get(i));
			ret[i] = new File(scriptURL.toURI());
		}
		return ret;
	}
	
	/**
	 * @param variable the name of the variable
	 * @param type the type of the declared variable
	 * @return a typed variable
	 */
	public <T> T get(String variable, Class<T> type){
		return (T) getVariables().get(variable);
	}
}
